import asyncio
import logging
import importlib
import signal
from contextlib import contextmanager


def get_module(module_path):
    module_path, _, child_name = module_path.rpartition('.')

    module = importlib.import_module(module_path)
    child = getattr(module, child_name)

    return module, child


class InterruptBumper(object):
    def __init__(self, attempts):
        self.attempts = attempts

    def __enter__(self):
        self.signal_received = ()
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        self.signal_received = (sig, frame)

        self.attempts -= 1
        logging.warn(f'Bumping interrupts. [remains={self.attempts}]')

        if not self.attempts:
            logging.warn('max attempts reached stop.')
            self.old_handler(*self.signal_received)

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received:
            self.old_handler(*self.signal_received)


@contextmanager
def simple_eventloop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    yield loop

    loop.close()



import asyncio
import logging
import signal
from contextlib import asynccontextmanager
from contextlib import suppress

import uvloop

from metropolis.core.utils import get_module
from metropolis.core.utils import simple_eventloop
from metropolis.core.driver import NatsDriver


WORKER_CONTROL_SIGNAL_START = 'START'
WORKER_CONTROL_SIGNAL_STOP = 'STOP'


def set_logger(conf):
    log_level = getattr(logging, conf.LOG_LEVEL, 'WARNING')
    logging.basicConfig(format=conf.LOG_FORMAT, level=log_level)


class Worker(object):
    conf = None

    # nats driver
    _driver = None

    # worker object's eventloop
    _loop = None

    # aio queue for worker lifecycle control
    _queue = None

    def __init__(self, conf, loop=None):
        self.conf = conf

        set_logger(self.conf)
        logging.info('Initialize application')

        logging.debug('Setup driver')
        _, serializer = get_module(conf.SERIALIZER_CLASS)
        self._driver = NatsDriver(
            urls=self.conf.NATS_URL.split(','),
            serializer=serializer)

        logging.debug('Setup uvloop')
        if self.conf.UVLOOP_ENABLED:
            uvloop.install()

        logging.debug('Prepare eventloop')
        self._loop = loop or asyncio.get_event_loop()

        logging.debug('Generate worker')
        self._queue = asyncio.Queue()

        logging.debug('Set signal handler')
        self._handle_signal = self.create_signal_handler()
        signal.signal(signal.SIGINT, self.stop)
        signal.signal(signal.SIGTERM, self.stop)

    def stop(self, *args, **kwargs):
        self._queue.put_nowait(WORKER_CONTROL_SIGNAL_STOP)

    def create_signal_handler(self):
        async def _handle_signal(msg):
            worker_message = msg.data.decode()
            logging.debug(f'Got worker signal [signal={worker_message}]')

            await self._queue.put_nowait(worker_message)
        return _handle_signal

    @property
    def run_until_complete(self):
        return self._loop.run_until_complete

    @asynccontextmanager
    async def nats_driver(self, loop=None):
        # remove condition for performance
        loop = loop or self._loop
        nats = await self._driver.get_connection(loop)

        yield nats

        # Gracefully unsubscribe the subscription
        await self._driver.close()

    async def _run_in_loop(self):
        async with self.nats_driver() as nats:

            # Setup worker lifecycle handler
            if self.conf.CONTROL_LIFECYCLE_ENABLED:
                await nats.subscribe(self.conf.WORKER_NAME, cb=self._handle_signal)

            # Register tasks
            for task_spec in self.conf.TASKS:
                _, task_fn = get_module(task_spec['task'])
                callback = self._driver.create_task(task_fn)

                subscription_id = await nats.subscribe(
                    task_spec['subject'], queue=task_spec['queue'], cb=callback)

                logging.debug((
                    'Task is registered '
                    f'[subscription_id={subscription_id}]'
                    f'[subject={task_spec["subject"]}]'
                    f'[queue={task_spec["queue"]}]'
                    f'[task={task_fn.__name__}]'
                ))

            # wait for stop signal
            signal = WORKER_CONTROL_SIGNAL_START
            while signal != WORKER_CONTROL_SIGNAL_STOP:
                signal = await self._queue.get()

    def run(self):
        try:
            logging.info('Start - run worker')
            self._loop.run_until_complete(self._run_in_loop())

        except KeyboardInterrupt:
            logging.debug(f'Stop - send stop message to worker')
            self.stop()

        finally:
            logging.info('Stop - cancel pending eventloop tasks')
            pending_tasks = asyncio.Task.all_tasks()
            for task in pending_tasks:
                logging.debug((
                    f'[task={task.__class__.__name__}:{task.__hash__()}]'
                ))
                with suppress(asyncio.CancelledError):
                    self._loop.run_until_complete(task)
                    logging.debug(f'{task.__hash__()} [done={task.done()}]')

            logging.info('Stop - close eventloop')
            self._loop.close()

        logging.info('Bye')

    async def async_request(self, name, payload, loop=None):
        async with self.nats_driver(loop) as nats:
            res = await nats.request(name, payload)
            return res

    async def async_publish(self, name, payload, loop=None):
        async with self.nats_driver(loop) as nats:
            await nats.publish(name, payload)

    def request(self, name, payload):
        with simple_eventloop() as loop:
            response = loop.run_until_complete(
                self.async_request(name, payload, loop))
            return response

    def publish(self, name, payload):
        with simple_eventloop() as loop:
            response = loop.run_until_complete(
                self.async_publish(name, payload, loop))
            return response



import ujson


class DefaultMessageSerializer(object):
    @staticmethod
    def serialize(msg):
        return msg.encode()

    @staticmethod
    def deserialize(msg):
        return msg.decode()


class JsonMessageSerializer(object):
    @staticmethod
    def serialize(msg):
        return ujson.dumps(msg).encode()

    @staticmethod
    def deserialize(msg):
        return ujson.loads(msg.decode())



import logging
import time

from nats.aio.client import Client

from metropolis.core.utils import InterruptBumper


class NatsDriver(object):
    nats = None
    serializer = None

    def __init__(self, urls, serializer):
        self.urls = urls
        self.serializer = serializer

    async def get_connection(self, loop):
        self.nats = Client()

        await self.nats.connect(
            servers=self.urls,
            loop=loop,
            error_cb=self.on_error,
            disconnected_cb=self.on_disconnected,
            closed_cb=self.on_closed,
            reconnected_cb=self.on_reconnected
        )

        return self.nats

    @staticmethod
    async def on_error(exception):
        logging.error(f'{exception}')

    @staticmethod
    async def on_disconnected():
        logging.info(f'disconnected')

    @staticmethod
    async def on_closed():
        logging.info(f'closed')

    @staticmethod
    async def on_reconnected():
        logging.info(f'reconnected')

    def create_task(self, task_fn):
        logging.debug(f'Create task [task_fn={task_fn.__name__}]')

        async def run_task(msg):
            if self.nats.is_draining:
                logging.debug('Connection is draining')
                raise Exception('draining')

            logging.info((
                'Received message. '
                f'[subject={msg.subject}][fn={task_fn.__name__}]'
                f'[from={msg.reply}]'
            ))

            try:
                with InterruptBumper(attempts=3):
                    now = time.perf_counter()
                    data = self.serializer.deserialize(msg.data)

                    try:
                        ret = task_fn(**data)
                        code = 200

                    except Exception as e:
                        ret = str(e)
                        code = 500

                    if msg.reply:
                        response_data = self.serializer.serialize({
                            'code': code,
                            'data': ret
                        })

                        await self.nats.publish(msg.reply, response_data)

                    elapsed = (time.perf_counter() - now) * 1000

                    logging.info((
                        'Task finished. '
                        f'[subject={msg.subject}][fn={task_fn.__name__}]'
                        f'[elapsed={elapsed:.3f}ms]'
                    ))

            except KeyboardInterrupt:
                await self.nats.publish(
                    msg.reply, 'KeyboardInterrupt'.encode())
                raise

        return run_task

    async def close(self):
        logging.debug('Drain subscriptions')

        await self.nats.flush()
        await self.nats.drain()
        await self.nats.close()



from sanic import Sanic
from sanic.response import json

from metropolis.core.driver import NatsDriver
from metropolis.core.utils import get_module


class WorkerHttpProxy(object):
    conf = None
    app = None
    driver = None
    nats = None

    def __init__(self, settings):
        self.conf = settings

        _, serializer = get_module(self.conf.SERIALIZER_CLASS)
        self.driver = NatsDriver(
            urls=self.conf.NATS_URL.split(','),
            serializer=serializer)

        self.app = Sanic()
        self.app.listener('before_server_start')(self.setup)
        self.app.route('/_routes/', methods=['GET'])(self.get_routes)
        self.app.route('/<path:[^/].*?>', methods=['GET'])(self.resolve_message)

    async def setup(self, app, loop):
        """ Bind worker with sanic application eventloop
        """

        self.nats = await self.driver.get_connection(loop)

    def serialize_request_to_nats_message(self, request, path: str) -> (str, str):
        """Resolve path to nats topic, messages

        Request path convention
        - topic: {path.replace('/', '.'}.{method}.{param['_worker']}
        - message: GET params | POST body

        :Params
            - request: sanic request
            - path <str>: path of url

        :Returns
            <(str, str)>: tuple of topic, message
        """

        nats_route = '.'.join(item for item in (
            path.replace('/', '.'),
            request.method.lower(),
            request.args.get('_worker', '')
        ) if item)

        return (nats_route, request.args)

    async def get_routes(self, request):
        self.nats
        return json({})

    async def resolve_message(self, request, path: str):
        (route, body) = self.serialize_request_to_nats_message(request, path)

        # data transport
        message = self.driver.serializer.serialize(body)
        worker_response = await self.nats.request(route, message)

        worker_response_dict = self.driver.serializer.deserialize(
            worker_response.data)
        response_data = worker_response_dict['data']

        return json(response_data, status=worker_response_dict['code'])



import setuptools

setuptools.setup(
    name="metropolis",
    version="0.1.0",
    license='MIT',
    author="ashon lee",
    author_email="ashon8813@gmail.com",
    description="Microservice gateway using Nats",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/ashon/metropolis",
    packages=setuptools.find_packages(),
    install_requires=open('requirements.txt').readlines(),
    python_requires='>=3',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)
